@@ -22,16 +22,16 @@
  * means only the basic inode info is checked, no fork checks.
  */
 static int
-check_aginode_block(xfs_mount_t	*mp,
-			xfs_agnumber_t	agno,
-			xfs_agblock_t	agbno)
+check_aginode_block(
+	xfs_mount_t		*mp,
+	xfs_agnumber_t		agno,
+	xfs_agblock_t		agbno)
 {
-
-	xfs_dinode_t	*dino_p;
-	int		i;
-	int		cnt = 0;
-	struct xfs_buf	*bp;
-	int		error;
+	struct xfs_dinode	*dino_p;
+	int			i;
+	int			cnt = 0;
+	struct xfs_buf		*bp;
+	int			error;
 
 	/*
 	 * it's ok to read these possible inode blocks in one at
@@ -229,8 +229,7 @@ verify_inode_chunk(xfs_mount_t		*mp,
 		/*
 		 * ok, put the record into the tree, if no conflict.
 		 */
-		if (find_uncertain_inode_rec(agno,
-				XFS_AGB_TO_AGINO(mp, start_agbno)))
+		if (find_inode_rec(mp, agno, XFS_AGB_TO_AGINO(mp, start_agbno)))
 			return(0);
 
 		start_agino = XFS_AGB_TO_AGINO(mp, start_agbno);
@@ -598,7 +597,7 @@ process_inode_chunk(
 	xfs_ino_t		parent;
 	ino_tree_node_t		*ino_rec;
 	struct xfs_buf		**bplist;
-	xfs_dinode_t		*dino;
+	struct xfs_dinode	*dino;
 	int			icnt;
 	int			status;
 	int			bp_found;
@@ -1135,6 +1134,7 @@ check_uncertain_aginodes(xfs_mount_t *mp, xfs_agnumber_t agno)
 	xfs_agino_t		i;
 	xfs_agino_t		agino;
 	int			got_some;
+	struct xfs_perag	*pag;
 
 	nrec = NULL;
 	got_some = 0;
@@ -1158,6 +1158,7 @@ check_uncertain_aginodes(xfs_mount_t *mp, xfs_agnumber_t agno)
 
 	do_warn(_("found inodes not in the inode allocation tree\n"));
 
+	pag = libxfs_perag_get(mp, agno);
 	do {
 		/*
 		 * check every confirmed (which in this case means
@@ -1169,7 +1170,7 @@ check_uncertain_aginodes(xfs_mount_t *mp, xfs_agnumber_t agno)
 
 			agino = i + irec->ino_startnum;
 
-			if (!libxfs_verify_agino(mp, agno, agino))
+			if (!libxfs_verify_agino(pag, agino))
 				continue;
 
 			if (nrec != NULL && nrec->ino_startnum <= agino &&
@@ -1178,7 +1179,7 @@ check_uncertain_aginodes(xfs_mount_t *mp, xfs_agnumber_t agno)
 				continue;
 
 			if ((nrec = find_inode_rec(mp, agno, agino)) == NULL)
-				if (libxfs_verify_agino(mp, agno, agino))
+				if (libxfs_verify_agino(pag, agino))
 					if (verify_aginode_chunk(mp, agno,
 							agino, &start))
 						got_some = 1;
@@ -1189,6 +1190,7 @@ check_uncertain_aginodes(xfs_mount_t *mp, xfs_agnumber_t agno)
 
 		irec = findfirst_uncertain_inode_rec(agno);
 	} while (irec != NULL);
+	libxfs_perag_put(pag);
 
 	if (got_some)
 		do_warn(_("found inodes not in the inode allocation tree\n"));
@@ -1229,6 +1231,7 @@ process_uncertain_aginodes(xfs_mount_t *mp, xfs_agnumber_t agno)
 	int			cnt;
 	int			got_some;
 	struct xfs_ino_geometry	*igeo = M_IGEO(mp);
+	struct xfs_perag	*pag;
 
 #ifdef XR_INODE_TRACE
 	fprintf(stderr, "in process_uncertain_aginodes, agno = %d\n", agno);
@@ -1243,6 +1246,7 @@ process_uncertain_aginodes(xfs_mount_t *mp, xfs_agnumber_t agno)
 
 	nrec = NULL;
 
+	pag = libxfs_perag_get(mp, agno);
 	do  {
 		/*
 		 * check every confirmed inode
@@ -1260,7 +1264,7 @@ process_uncertain_aginodes(xfs_mount_t *mp, xfs_agnumber_t agno)
 			 * good tree), bad inode numbers, and inode numbers
 			 * pointing to bogus inodes
 			 */
-			if (!libxfs_verify_agino(mp, agno, agino))
+			if (!libxfs_verify_agino(pag, agino))
 				continue;
 
 			if (nrec != NULL && nrec->ino_startnum <= agino &&
@@ -1304,6 +1308,7 @@ process_uncertain_aginodes(xfs_mount_t *mp, xfs_agnumber_t agno)
 
 		irec = findfirst_uncertain_inode_rec(agno);
 	} while (irec != NULL);
+	libxfs_perag_put(pag);
 
 	if (got_some)
 		do_warn(_("found inodes not in the inode allocation tree\n"));
